!---------------------------------------------------------------
! Copyright (C) 2009-2015 GFD Dennou Club. All rights reserved.
!---------------------------------------------------------------

module cloud_basic
! CReSS の雲物理過程
! 本関数内では, 値の大きな物理量のべき乗計算が出てきた際,
! 桁落ちを防ぐため意図的に exp(log()) という計算を行って
! 回避する.
! たとえば, 任意の物理量 a**b の場合, exp(b*log(a)) がそれに相当する.
! なぜなら, a**b=exp(log(a**b))=exp(b*log(a)) だからである.
! また, おおきな物理量 a*b の場合も, exp(log(a)+log(b)) としている.
! これらのおおきな物理量が割り算に入っている場合,
! exp(-a) とすることで割り算を回避する.
  use Thermo_Function
  use Thermo_Const
  use special_function
  use Derivation
  use Cloud_Const
  use Math_Const
  use Phys_Const

contains

real function Lv( temp )
! 水 -> 水蒸気での潜熱加熱
  use Thermo_Const
  implicit none
  real, intent(in) :: temp  ! 温度 [K]

  call ptcheck( temp, 'Lv' )

  Lv=2.50078e6*(ti0/temp)**(0.167+3.67e-4*temp)

  return

end function Lv

real function Ls( temp )
! 氷 -> 水蒸気での潜熱加熱
  use Thermo_Const
  implicit none
  real, intent(in) :: temp ! 温度 [K]

  call ptcheck( temp, 'Ls' )

  Ls=2.834e6+100.0*(temp-ti0)

  return

end function Ls

real function Lf( temp )
! 水 -> 氷での融解熱
  use Thermo_Const
  implicit none
  real, intent(in) :: temp ! 温度 [K]

  call ptcheck( temp,'Lf' )

  Lf=3.34e5+2500.0*(temp-ti0)

  return

end function Lf

real function nu_air( temp, pres )
! 空気の動粘性係数
  use Thermo_Const
  implicit none
  real, intent(in) :: temp ! 温度 [K]
  real, intent(in) :: pres ! 気圧 [Pa]

  call ptcheck( temp, 'nu air' )

  nu_air=1.328e-5*(p00/pres)*(temp/ti0)**1.754

  return

end function nu_air

real function mu_air( temp, pres )
! 空気の粘性係数
  implicit none
  real, intent(in) :: temp ! 温度 [K]
  real, intent(in) :: pres ! 気圧 [Pa]
  real :: rho

  call ptcheck( temp, 'mu air' )

  rho=TP_2_rho( temp, pres )

  mu_air=nu_air( temp, pres )*rho

  return

end function mu_air

real function Dv( temp, pres )
! 水蒸気の拡散係数
  use Thermo_Const
  implicit none
  real, intent(in) :: temp ! 温度 [K]
  real, intent(in) :: pres ! 気圧 [Pa]

  call ptcheck( temp, 'Dv' )

  Dv=2.23e-5*(p00/pres)*(temp/ti0)**1.81

  return

end function Dv

real function qvss( temp, pres )
  use Thermo_Const
  implicit none
  real, intent(in) :: temp
  real, intent(in) :: pres
  real :: eps
  real :: a, b, c, tmp

  call ptcheck( temp, 'qvss' )

  a=610.78
  b=17.269
  c=35.86
  eps=Rd/Rv
  tmp=b*(temp-t0)/(temp-c)

  qvss=eps*a*exp(tmp)/pres

  return

end function qvss

real function qvsi( temp, pres )
  use Thermo_Const
  implicit none
  real, intent(in) :: temp
  real, intent(in) :: pres
  real :: eps
  real :: a, b, c, tmp

  call ptcheck( temp, 'qvsi' )

  a=610.78
  b=21.875
  c=7.86
  eps=Rd/Rv
  tmp=b*(temp-t0)/(temp-c)

  qvsi=eps*a*exp(tmp)/pres

  return

end function qvsi

!---------------------------------
! 以下, プライベート関数
!---------------------------------

real function lamx( types, qx, rhob, Nx )
! 凝結物質の指数関数分布における傾きを計算する.
! 数濃度がない場合は, y 切片を固定値で仮定して計算する.
  use Math_Const
  use Cloud_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度
  real :: tmpa

  if(qx>=qthres)then
     if(present(Nx).and.types(1:1)/='r')then  ! 'r' は Nx を使わないので, 場合分けしない.
        tmpa=pi*Nx/rhob/qx
        select case (types(1:1))
        case ('s')
           lamx=exp((1.0/3.0)*log(tmpa*rhos))
        case ('g')
           lamx=exp((1.0/3.0)*log(tmpa*rhog))
        case('r')
           write(*,*) "*** ERROR (lamx) *** :"
           write(*,*) "type = 'r' => Disable using optional ARG."
           write(*,*) "STOP"
           stop
        end select
     else
        tmpa=pi/rhob/qx
        select case (types(1:1))
        case ('r')
           lamx=exp(0.25*log(tmpa*rhow*nr0))
        case ('s')
           lamx=exp(0.25*log(tmpa*rhos*ns0))
        case ('g')
           lamx=exp(0.25*log(tmpa*rhog*ng0))
        end select
     end if
  else
     lamx=0.0
  end if

  return

end function lamx

real function nx0( types, qx, rhob, Nx )
! 凝結物質の指数関数分布における y 切片を計算する.
! 数濃度がない場合は, y 切片を固定値で仮定して計算する.
  use Math_Const
  use Cloud_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度
  real :: tmpa

  if(qx>=qthres)then
     if(present(Nx).and.types(1:1)/='r')then  ! 'r' は Nx を使わないので, 場合分けしない.
        tmpa=pi*Nx/rhob/qx
        select case (types(1:1))
        case ('s')
           nx0=exp(log(Nx)+(1.0/3.0)*log(tmpa*rhos))
        case ('g')
           nx0=exp(log(Nx)+(1.0/3.0)*log(tmpa*rhog))
        case('r')
           write(*,*) "*** ERROR (nx0) *** :"
           write(*,*) "type = 'r' => Disable using optional ARG."
           write(*,*) "STOP"
           stop
        end select
     else
        select case (types(1:1))
        case ('r')
           nx0=nr0
        case ('s')
           nx0=ns0
        case ('g')
           nx0=ng0
        end select
     end if
  else
     nx0=0.0
  end if

  return

end function nx0


real function Knudsen( temp, pres )
! Knudsen 数を計算する.
  use Cloud_Const
  use Thermo_Const
  implicit none
  real, intent(in) :: temp   ! 温度 [k]
  real, intent(in) :: pres   ! 圧力 [Pa]

  call ptcheck( temp, 'Knudsen' )

  Knudsen=MFP0*p00*temp/(tref*Raero*pres)

  return

end function


real function Daero( temp, pres )
! エアロゾルの拡散係数を計算する.
! 現在, 雲水の温度と周囲の温度は同じとしている.
  use Cloud_Const
  use Thermo_Const
  use Phys_Const
  use Math_Const
  implicit none
  real, intent(in) :: temp   ! 温度 [k]
  real, intent(in) :: pres   ! 圧力 [Pa]
  real :: Tcl

  call ptcheck( temp, 'Daero' )

  Tcl=temp
  Daero=kB*Tcl*(1.0+Knudsen( temp, pres ))/(6.0*pi*Raero*mu_air( temp, pres ))

  return

end function


real function Naero( temp )
! 接触凍結核の数濃度を計算する.
! ここでは雲粒の温度はその場の大気の温度として計算している.
  use Cloud_Const
  use Thermo_Const
  use Phys_Const
  use Math_Const
  implicit none
  real, intent(in) :: temp   ! 温度 [k]

  call ptcheck( temp, 'Naero' )

  if(270.16>temp)then
     Naero=Na0*((270.16-temp)**(1.3))
  else
     Naero=0.0
  end if

  return

end function Naero


real function Dbc( qc, rhob )
! 単分散分布の雲水平均直径を計算する. (4.51)
  use Cloud_Const
  use Thermo_Const
  use Math_Const
  implicit none
  real, intent(in) :: qc     ! 雲水混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場密度 [kg/m3]
  real :: tmp

  tmp=6.0*rhob*qc/(pi*rhow*Nc)
  Dbc=exp((1.0/3.0)*log(tmp))

  return

end function Dbc


real function Dbi( qi, rhob, Ni )
! 単分散分布の雲氷平均直径を計算する. (4.52)
! Ni が陽に与えられない場合, temp, pres, qv から診断する.
  use Cloud_Const
  use Thermo_Const
  use Math_Const
  implicit none
  real, intent(in) :: qi     ! 雲氷混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場密度 [kg/m3]
  real, intent(in), optional :: Ni     ! 雲氷の数濃度 [1/m3]
  real :: tmp, tmpNi

  if(present(Ni))then
     tmpNi=Ni
  else
     tmpNi=Ni_calc( qi )
  end if

  if(tmpNi>0.0)then
     tmp=6.0*rhob*qi/(pi*rhoi*tmpNi)
     Dbi=exp((1.0/3.0)*log(tmp))
  else
     Dbi=0.0
  end if

  return

end function Dbi


real function Ni_calc( qi )
! diagni.f90 を用いて, 氷晶の数濃度を計算する.
  use Cloud_Const
  use Thermo_Const
  implicit none
  real, intent(in) :: qi     ! 雲氷の混合比 [kg/kg]

  Ni_calc=4.0*qi/(3.0*mi0max)  

  return

end function Ni_calc


!real function Ni_calc( temp, pres, qv )  これは用いない.
!! (4.75) を用いて, 氷晶の数濃度を計算する.
!  use Cloud_Const
!  use Thermo_Const
!  implicit none
!  real, intent(in) :: temp   ! 温度 [K]
!  real, intent(in) :: pres   ! 圧力 [Pa]
!  real, intent(in) :: qv     ! 水蒸気混合比 [kg/kg]
!  real :: Si, temps, SSi, Swi, tmpa
!  
!  call ptcheck( temp )
!
!  Ni_calc=0.0
!
!  if(temp<=ti0)then
!     temps=ti0-temp
!     Si=qv/qvsi( temp, pres )
!     SSi=Si-1.0
!     Swi=qvss( temp, pres )/qvsi( temp, pres )
!     if(SSi>0.0)then
!        if(Swi>1.0)then
!           tmpa=(Si-1.0)/(Swi-1.0)
!           if(temp>=t0-5.0)then
!              Ni_calc=Ni01*exp(FLT2*temps)*exp(HVB*log(tmpa))
!           else
!              Ni_calc=Ni02*exp(Fea1*SSi-Feb1)
!           end if
!        end if
!     end if
!  end if
!
!  return
!
!end function Ni_calc


real function Koenig1( temp )
  implicit none
  real, intent(in) :: temp   ! 温度 [K]
  real :: tc
  real, dimension(32) :: param
  integer :: i

  call ptcheck( temp, 'Koenig1' )

  param=(/0.0,  &
  &       7.939e-11,  &
  &       7.841e-10,  &
  &       3.369e-9,  &
  &       4.336e-9,  &
  &       5.285e-9,  &
  &       3.728e-9,  &
  &       1.852e-9,  &
  &       2.991e-10,  &
  &       4.248e-10,  &
  &       7.434e-10,  &
  &       1.812e-9,  &
  &       4.394e-9,  &
  &       9.145e-9,  &
  &       1.725e-10,  &
  &       3.348e-8,  &
  &       1.725e-8,  &
  &       9.175e-9,  &
  &       4.412e-9,  &
  &       2.252e-9,  &
  &       9.115e-10,  &
  &       4.876e-10,  &
  &       3.473e-10,  &
  &       4.758e-10,  &
  &       6.306e-10,  &
  &       8.573e-10,  &
  &       7.868e-10,  &
  &       7.192e-10,  &
  &       6.153e-10,  &
  &       5.956e-10,  &
  &       5.333e-10,  &
  &       4.834e-10 /)

  tc=temp-t0
  Koenig1=0.0

  if(tc<0.0)then
     do i=1,32
        if(temp>=-real(i).and.temp<-real(i-1))then
           Koenig1=param(i)
           exit
        end if
     end do
  end if

  return

end function Koenig1


real function Koenig2( temp )
  implicit none
  real, intent(in) :: temp   ! 温度 [K]
  real :: tc
  real, dimension(32) :: param
  integer :: i

  call ptcheck( temp, 'Koenig2' )

  param=(/0.0,  &
  &       4.006e-1,  &
  &       4.831e-1,  &
  &       5.320e-1,  &
  &       5.307e-1,  &
  &       5.319e-1,  &
  &       5.249e-1,  &
  &       4.888e-1,  &
  &       3.894e-1,  &
  &       4.047e-1,  &
  &       4.318e-1,  &
  &       4.771e-1,  &
  &       5.183e-1,  &
  &       5.463e-1,  &
  &       5.651e-1,  &
  &       5.813e-1,  &
  &       5.655e-1,  &
  &       5.478e-1,  &
  &       5.203e-1,  &
  &       4.906e-1,  &
  &       4.447e-1,  &
  &       4.126e-1,  &
  &       3.960e-1,  &
  &       4.149e-1,  &
  &       4.320e-1,  &
  &       4.506e-1,  &
  &       4.483e-1,  &
  &       4.460e-1,  &
  &       4.433e-1,  &
  &       4.413e-1,  &
  &       4.382e-1,  &
  &       4.316e-1 /)

  tc=temp-t0
  Koenig2=0.0

  if(tc<0.0)then
     do i=1,32
        if(temp>=-real(i).and.temp<-real(i-1))then
           Koenig2=param(i)
           exit
        end if
     end do
  end if

  return

end function Koenig2


real function Exy( typexy, temp, pres, qc, qx, rhob, rho0, Nx )
! 衝突併合における捕捉率を求める.
  use Cloud_Const
  implicit none
  character(2), intent(in) :: typexy  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰, 'c' = 雲水, 'i' = 雲氷
  real, intent(in), optional :: temp ! 温度 [K]
  real, intent(in), optional :: pres ! 気圧 [Pa]
  real, intent(in), optional :: qc     ! 雲水の混合比 [kg/kg]
  real, intent(in), optional :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in), optional :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in), optional :: rho0   ! 地上における基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度
  real :: tmp_Stk

  if(typexy(1:1)=='c')then
     if(qc>=qthres.and.qx>=qthres)then
        if(present(Nx))then
           tmp_Stk=Stk( typexy(2:2), temp, pres, qc, qx, rhob, rho0, Nx )
        else
           tmp_Stk=Stk( typexy(2:2), temp, pres, qc, qx, rhob, rho0 )
        end if
        Exy=(tmp_Stk/(tmp_Stk+0.5))**2
     else
        Exy=0.0
     end if
  else if(typexy(1:1)=='r')then
     Exy=1.0
  else if(typexy(1:1)=='i')then
     if(typexy(2:2)=='r')then  ! [CReSS orig] Eis and Eig = 0.1
        Exy=1.0
     else
        Exy=0.1
     end if
  else if(typexy(1:1)=='s')then
     if(typexy(2:2)=='g')then
        Exy=0.001
     else
        Exy=1.0
     end if
  else
     write(*,*) "ERROR (Exy) : typexy(1:1) is not valid ("//typexy(1:1)//")."
  end if

end function Exy


real function Stk( types, temp, pres, qc, qx, rhob, rho0, Nx )
! ストークス比を計算する. (4.110)
  use Cloud_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: temp ! 温度 [K]
  real, intent(in) :: pres ! 気圧 [Pa]
  real, intent(in) :: qc     ! 雲水の混合比 [kg/kg]
  real, intent(in) :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in) :: rho0   ! 地上における基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度

  real :: tmpDbc, Dy, Uby, mu

  call ptcheck( temp, 'Stk' )

  if(qc>=qthres.and.qx>=qthres)then
     tmpDbc=Dbc( qc, rhob )
     if(present(Nx))then
        Uby=UhxN( types(1:1), qx, rhob, rho0, Nx )
        Dy=1.0/lamx( types, qx, rhob, Nx )
     else
        Uby=UhxN( types(1:1), qx, rhob, rho0 )
        Dy=1.0/lamx( types, qx, rhob )
     end if
     mu=mu_air( temp, pres )

     Stk=tmpDbc*tmpDbc*rhow*Uby/(9.0*mu*Dy)

  else

     Stk=0.0

  end if

  return

end function Stk


real function Category_rs( qr, qs, rhob, Ns )
! 雨と雪の衝突後の分配係数を計算する. (4.113)
  use Cloud_Const
  use Thermo_Const
  implicit none
  real, intent(in) :: qr  ! 雨の混合比 [kg/kg]
  real, intent(in) :: qs  ! 雪の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in), optional :: Ns  ! 雪の数濃度 [1/m3]
  real :: lamdas, lamdar, mbr, mbs

  if(qr>=qthres.and.qs>=qthres)then
     if(present(Ns))then
        lamdas=lamx( 's', qs, rhob, Ns )
     else
        lamdas=lamx( 's', qs, rhob )
     end if

     lamdar=lamx( 'r', qr, rhob )
     mbr=rhow*exp(3.0*(log(4.0)-log(lamdar)))
     mbs=rhos*exp(3.0*(log(4.0)-log(lamdas)))

     Category_rs=mbs*mbs/(mbs*mbs+mbr*mbr)

  else

     Category_rs=0.0

  end if

  return

end function Category_rs


real function Gw( temp, pres, rhob )
  use Cloud_Const
  use Math_Const
  use Thermo_Const
  implicit none
  real, intent(in) :: temp  ! 温度 [K]
  real, intent(in) :: pres  ! 圧力 [Pa]
  real, intent(in) :: rhob  ! 基準密度 [kg/m3]
  real :: tmpa, tmpb, tmpc, tmpd

  call ptcheck( temp, 'Gw' )

  tmpa=Lv(temp)
  tmpb=kappaa*Rv*temp*temp
  tmpc=rhob*qvss(temp,pres)*Dv(temp,pres)
  tmpd=tmpa*tmpa/tmpb+1.0/tmpc
  Gw=1.0/tmpd

  return

end function Gw


real function Gi( temp, pres, rhob )
  use Cloud_Const
  use Math_Const
  use Thermo_Const
  implicit none
  real, intent(in) :: temp  ! 温度 [K]
  real, intent(in) :: pres  ! 圧力 [Pa]
  real, intent(in) :: rhob  ! 基準密度 [kg/m3]
  real :: tmpa, tmpb, tmpc, tmpd

  call ptcheck( temp, 'Gi' )

  tmpa=Ls(temp)
  tmpb=kappaa*Rv*temp*temp
  tmpc=rhob*qvsi(temp,pres)*Dv(temp,pres)
  tmpd=tmpa*tmpa/tmpb+1.0/tmpc
  Gi=1.0/tmpd

  return

end function Gi


real function VENTx( types, temp, pres, qx, rhob, rho0, Nx )
  use Cloud_Const
  use Math_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: temp  ! 温度 [K]
  real, intent(in) :: pres  ! 圧力 [Pa]
  real, intent(in) :: qx    ! types の混合比 [kg/kg]
  real, intent(in) :: rhob  ! 基準密度 [kg/m3]
  real, intent(in) :: rho0  ! 地表面での基準密度 [kg/m3]
  real, intent(in), optional :: Nx    ! types の数濃度 [1/m3]
  real :: tmp, tmpa, tmpb, tmpc, aux, bux, ny0, lamdax

  call ptcheck( temp, 'VENTx' )

  if(qx>=qthres)then
     select case (types(1:1))
     case ('r')
        aux=aur
        bux=bur

     case ('s')
        aux=aus
        bux=bus

     case ('g')
        aux=aug
        bux=bug

     end select

     if(present(Nx))then
        lamdax=lamx( types(1:1), qx, rhob, Nx )
        ny0=nx0( types(1:1), qx, rhob, Nx )
     else
        lamdax=lamx( types(1:1), qx, rhob )
        ny0=nx0( types(1:1), qx, rhob )
     end if

     tmp=0.5*(5.0+bux)
     tmpa=0.78*exp(-2.0*log(lamdax))
     tmpb=0.31*(Sc**(1.0/3.0))*sqrt(aux/nu_air( temp, pres ))  &
  &       *gamma_func(tmp)*sqrt(sqrt(rho0/rhob))
     tmpc=exp(-tmp*log(lamdax))
!     VENTx=rhob*ny0*(tmpa+tmpb*tmpc)  ! [CReSS orig]
     VENTx=ny0*(tmpa+tmpb*tmpc)

  else

     VENTx=0.0

  end if

  return

end function VENTx


real function UbxN( types, qx, rhob, rho0, Nx )
! 数濃度の重みをかけたカテゴリー x の終端落下速度を計算する.
! 数濃度がない場合は, y 切片を固定値で仮定して計算する.
  use Math_Const
  use Cloud_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in) :: rho0   ! 地上における基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度
  real :: tmpa, tmpb, tmpc, lamxx

  if(qx>=qthres)then
     if(present(Nx))then
        lamxx=lamx( types(1:1), qx, rhob, Nx )
     else
        lamxx=lamx( types(1:1), qx, rhob )
     end if

     select case (types(1:1))  ! 'r' は Nx を使わないので, 場合分けしない.
     case ('r')
        tmpa=aur*gamma_func(1.0+bur)
        tmpb=exp(-bur*log(lamxx))
        tmpc=(rho0/rhob)**(cur)
     case ('s')
        tmpa=aus*gamma_func(1.0+bus)
        tmpb=exp(-bus*log(lamxx))
        tmpc=(rho0/rhob)**(cus)
     case ('g')
        tmpa=aug*gamma_func(1.0+bug)
        tmpb=exp(-bug*log(lamxx))
        tmpc=(rho0/rhob)**(cug)
     end select

     UbxN=tmpa*tmpc*tmpb

  else

     UbxN=0.0

  end if

  return

end function UbxN


subroutine ptcheck( temp, rname )
  implicit none
  real, intent(in) :: temp  ! 温度 [K]
  character(*), intent(in), optional :: rname  ! routine name

  if(temp>1000.0)then
     if(present(rname))then
        write(*,*) "ERROR ptcheck ("//trim(rname)//')', temp
     else
        write(*,*) "ERROR ptcheck", temp
     end if

     stop
  end if

end subroutine ptcheck


real function UhxN( types, qx, rhob, rho0, Nx )
! 重量の重みをかけたカテゴリー x の終端落下速度を計算する.
! 数濃度がない場合は, y 切片を固定値で仮定して計算する.
  use Math_Const
  use Cloud_Const
  use Thermo_Const
  implicit none
  character(1), intent(in) :: types  ! 凝結物質の種類 'r' = 雨, 's' = 雪, 'g' = 霰
  real, intent(in) :: qx     ! 凝結物質の混合比 [kg/kg]
  real, intent(in) :: rhob   ! 基本場の密度 [kg/m3]
  real, intent(in) :: rho0   ! 地上における基本場の密度 [kg/m3]
  real, intent(in), optional :: Nx  ! qx という凝結物質の数濃度
  real :: tmpa, tmpb, tmpc, lamxx

  if(qx>=qthres)then

     if(present(Nx))then
        lamxx=lamx( types(1:1), qx, rhob, Nx )
     else
        lamxx=lamx( types(1:1), qx, rhob )
     end if

     select case (types(1:1))  ! 'r' は Nx を使わないので, 場合分けしない.
     case ('r')
        tmpa=aur*gamma_func(4.0+bur)
        tmpb=exp(-(log(6.0)+bur*log(lamxx)))
        tmpc=(rho0/rhob)**(cur)
     case ('s')
        tmpa=aus*gamma_func(4.0+bus)
        tmpb=exp(-(log(6.0)+bus*log(lamxx)))
        tmpc=(rho0/rhob)**(cus)
     case ('g')
        tmpa=aug*gamma_func(4.0+bug)
        tmpb=exp(-(log(6.0)+bug*log(lamxx)))
        tmpc=(rho0/rhob)**(cug)
     end select

     UhxN=tmpa*tmpc*tmpb

  else

     UhxN=0.0

  end if

  return

end function UhxN


subroutine Moist_Sature_Adjust( pres, pt, qv, qc, err, ln )
  use Thermo_Const
  implicit none
  real, intent(in) :: pres ! 圧力 [Pa]
  real, intent(inout) :: pt   ! 温位 [K]
  real, intent(inout) :: qv   ! 水蒸気混合比 [kg kg-1]
  real, intent(inout) :: qc   ! 雲水混合比 [kg kg-1]
  real, intent(in), optional :: err  ! convergence limit [default = 1.0e-8]
  integer, intent(in), optional :: ln  ! ループ回数
                                       ! これが設定された場合, 相対誤差による
                                       ! ループ打ち切り判断は行わない.
  real :: tmppt, tmpqv, tmpqc, dqc, temp, gam, gam1, coe, pt1, qv1, qc1, err_max
  real :: tmp_err
  integer :: cn

  call ptcheck( thetaP_2_T( pt, pres ), 'Moist Sature Adjust' )

  if(qv<0.0.or.qc<0.0)then
     write(*,*) "ERROR : Argument in Moist_Sature_Adjust must be greater than zero."
     write(*,*) "STOP."
     stop
  end if

  if(present(err))then
     err_max=err
  else
     err_max=1.0e-8
  end if

  temp=thetaP_2_T( pt, pres )
  dqc=qv-qvss( temp, pres )
  gam=Lv( temp )/(Cpd*exner_func_dry( pres ))
  gam1=Lv( temp )/Cpd
  pt1=pt
  qv1=qv
  qc1=qc

  if(present(ln))then

     do cn=1,ln
        if(dqc>0.0.or.qc1>0.0)then
           coe=qvss( temp, pres )*17.269*(t0-35.86)/((temp-35.86)**2)
           tmppt=pt1+gam*(dqc)/(1.0+gam1*coe)
           tmpqv=qv1+(pt1-tmppt)/gam
           tmpqc=qv1+qc1-tmpqv
           if(tmpqc>0.0)then
              pt1=tmppt
              qv1=tmpqv
              qc1=tmpqc
           else if(tmpqc<=0.0)then
              pt1=pt1-gam*qc1
              qv1=qv1+qc1
              qc1=0.0
           end if
           temp=thetaP_2_T( pt1, pres )
           dqc=qv1-qvss( temp, pres )
        end if
     end do

  else

     do while (dqc>0.0.or.qc1>0.0)
        coe=qvss( temp, pres )*17.269*(t0-35.86)/((temp-35.86)**2)
        tmppt=pt1+gam*(dqc)/(1.0+gam1*coe)
        tmpqv=qv1+(pt1-tmppt)/gam
        tmpqc=qv1+qc1-tmpqv
        if(tmpqc>0.0)then
           pt1=tmppt
           qv1=tmpqv
           qc1=tmpqc
        else if(tmpqc<=0.0)then
           pt1=pt1-gam*qc1
           qv1=qv1+qc1
           qc1=0.0
        end if
        temp=thetaP_2_T( pt1, pres )
        dqc=qv1-qvss( temp, pres )

        tmp_err=abs(tmppt-pt1)
        if(err_max>tmp_err)then
           exit
        end if
     end do

  end if

  pt=pt1
  qv=qv1
  qc=qc1


end subroutine Moist_Sature_Adjust


subroutine Moist_Sature_Adjust_Ice( pres, pt, qv, qi, err, ln )
  use Thermo_Const
  implicit none
  real, intent(in) :: pres ! 圧力 [Pa]
  real, intent(inout) :: pt   ! 温位 [K]
  real, intent(inout) :: qv   ! 水蒸気混合比 [kg kg-1]
  real, intent(inout) :: qi   ! 雲氷混合比 [kg kg-1]
  real, intent(in), optional :: err  ! convergence limit [default = 1.0e-8]
  integer, intent(in), optional :: ln  ! ループ回数
                                       ! これが設定された場合, 相対誤差による
                                       ! ループ打ち切り判断は行わない.
  real :: tmppt, tmpqv, tmpqi, dqi, temp, gam, gam1, coe, pt1, qv1, qi1, err_max
  real :: tmp_err
  integer :: cn

  call ptcheck( thetaP_2_T( pt, pres ), 'Moist Sature Adjust Ice' )

  if(qv<0.0.or.qi<0.0)then
     write(*,*) "ERROR : Argument in Moist_Sature_Adjust must be greater than zero."
     write(*,*) "STOP."
     stop
  end if

  if(present(err))then
     err_max=err
  else
     err_max=1.0e-8
  end if

  temp=thetaP_2_T( pt, pres )
  dqi=qv-qvsi( temp, pres )
  gam=Lf( temp )/(Cpd*exner_func_dry( pres ))
  gam1=Lf( temp )/Cpd
  pt1=pt
  qv1=qv
  qi1=qi

  if(present(ln))then

     do cn=1,ln
        if(dqi>0.0.or.qi1>0.0.or.temp<ti0)then
           coe=qvsi( temp, pres )*21.875*(t0-7.86)/((temp-7.86)**2)
           tmppt=pt1+gam*(dqi)/(1.0+gam1*coe)
           tmpqv=qv1+(pt1-tmppt)/gam
           tmpqi=qv1+qi1-tmpqv
           if(tmpqi>0.0)then
              pt1=tmppt
              qv1=tmpqv
              qi1=tmpqi
           else if(tmpqi<=0.0)then
              pt1=pt1-gam*qi1
              qv1=qv1+qi1
              qi1=0.0
           end if
           temp=thetaP_2_T( pt1, pres )
           dqi=qv1-qvsi( temp, pres )
        end if
     end do

  else

     do while (dqi>0.0.or.qi1>0.0.or.temp<ti0)
        coe=qvsi( temp, pres )*21.875*(t0-7.86)/((temp-7.86)**2)
        tmppt=pt1+gam*(dqi)/(1.0+gam1*coe)
        tmpqv=qv1+(pt1-tmppt)/gam
        tmpqi=qv1+qi1-tmpqv
        if(tmpqi>0.0)then
           pt1=tmppt
           qv1=tmpqv
           qi1=tmpqi
        else if(tmpqi<=0.0)then
           pt1=pt1-gam*qi1
           qv1=qv1+qi1
           qi1=0.0
        end if
        temp=thetaP_2_T( pt1, pres )
        dqi=qv1-qvsi( temp, pres )

        tmp_err=abs(tmppt-pt1)
        if(err_max>tmp_err)then
           exit
        end if
     end do

  end if

  pt=pt1
  qv=qv1
  qi=qi1

end subroutine Moist_Sature_Adjust_Ice



end module
